import sqlite3
import os


class Table:
    def __init__(self, cursor: sqlite3.Cursor):
        self._cursor = cursor

    @staticmethod
    def _insert_query(table, data: dict):
        return "insert into {table} ({rows}) values ({values})".format(
            table=table,
            rows=','.join(data.keys()),
            values=','.join(['?' for _ in range(len(data))])
        ), tuple(data.values())


class LogTags(Table):
    table_name = 'log_tags'

    name = 'name'
    value = 'value'
    color = 'color'

    def __init__(self, cursor: sqlite3.Cursor):
        super().__init__(cursor=cursor)

    def insert(self, name: str, value: int, color: str):
        data = {
            self.name : name,
            self.value: value,
            self.color: color
        }

        query = self._insert_query(table=self.table_name, data=data)
        self._cursor.execute(*query)


class LogEntries(Table):
    table_name = 'log_entries'

    id = 'id'
    file_path = 'file_path'
    function_name = 'function_name'
    line_num = 'line_num'
    msg = 'msg'
    tag_name = 'tag_name'
    depth = 'depth'
    thread_id = 'thread_id'
    thread_name = 'thread_name'
    is_main_thread = 'is_main_thread'
    timestamp = 'timestamp'

    def __init__(self, cursor: sqlite3.Cursor):
        super().__init__(cursor=cursor)

    def insert(self, file_path: str, function_name: str, line_num: int, msg: str, tag_name: str, depth: int,
               thread_id: int, thread_name: str, is_main_thread: bool):
        data = {
            self.file_path     : file_path,
            self.function_name : function_name,
            self.line_num      : line_num,
            self.msg           : msg,
            self.tag_name      : tag_name,
            self.depth         : depth,
            self.thread_id     : thread_id,
            self.thread_name   : thread_name,
            self.is_main_thread: int(is_main_thread)
        }

        query = self._insert_query(table=self.table_name, data=data)
        self._cursor.execute(*query)


class LoggerSql:
    def __init__(self, db_file_name: str):
        _sql = sqlite3.connect(db_file_name, isolation_level=None, check_same_thread=False)
        self._cursor = _sql.cursor()
        self.log_tags = LogTags(cursor=self._cursor)
        self.log_entries = LogEntries(cursor=self._cursor)

    def create_database(self):
        with open(os.path.abspath(f'{os.path.dirname(__file__)}/logger.sql'), 'r') as sql:
            script = sql.read()
        self._cursor.executescript(script)

    def select(self, query: str, parameters: tuple = tuple()):
        results = self._cursor.execute(query, parameters).fetchall()
        description = self._cursor.description
        return SelectResult(
            columns=[desc[0] for desc in description],
            rows=results
        )


class SelectResult:
    def __init__(self, columns: list, rows: list):
        self.columns = columns
        self.rows = rows

    @property
    def dict(self):
        return dict(zip(self.columns, *self.rows))

    def iterate(self):
        for i in range(len(self.rows)):
            yield dict(zip(self.columns, self.rows[i]))



from abc import abstractmethod
from typing import List


class LogTag:
    def __init__(self, name: str, value: int, html_color: str = None):
        self.name = name
        self.alias = ''.join([c.lower() for c in name if c.isalpha()])
        self.value = value
        self.html_color = html_color


class LogTagTemplate:
    @property
    @abstractmethod
    def default(self) -> LogTag:
        raise NotImplementedError('Default log level definition required!')

    @property
    @abstractmethod
    def critical(self) -> LogTag:
        raise NotImplementedError('Critical log level definition required!')

    @classmethod
    def get_all(cls) -> List[LogTag]:
        return [ll for item in dir(cls) if isinstance(ll := getattr(cls, item), LogTag)]


class LogTags(LogTagTemplate):
    default = LogTag(
        name='Default',
        value=0,
        html_color='limegreen'
    )

    critical = LogTag(
        name='Critical',
        value=90,
        html_color='red'
    )



from typing import List, Dict, Optional, Tuple
import os
import shutil
import re
import uuid
from pathlib import Path
from datetime import datetime
from pygments import highlight
from pygments.lexers.python import PythonLexer
from pygments.formatters.html import HtmlFormatter
import asyncio
import htmlmin
from dblogging.sqlite.dal import LoggerSql, SelectResult
from dblogging.config import LogTag


class HtmlLogGenerator:
    def __init__(self):
        self._sql = None  # type: Optional[LoggerSql]
        self.python_css_file = Path(os.path.abspath(f'{os.path.dirname(__file__)}/python.css'))
        self.main_css_file = Path(os.path.abspath(f'{os.path.dirname(__file__)}/styles.css'))
        self.js_file = Path(os.path.abspath(f'{os.path.dirname(__file__)}/actions.js'))

    def generate(self, log_file: str, title: str = None, include_code: bool = True,
                 datetime_range: Tuple[datetime, datetime] = None, exclude_files: List[str] = None):
        """
        Generates an HTML file that display logs persisted by the logger in the SQLite DB. Consider the
        parameter definitions for ways to reduce the size of the output and control the content in the
        output.

        Args:
            log_file: Absolute path to the SQLite DB log file.
            title: Title of the log file. Default is "Log File".
            include_code: If ``False``, then the code blocks are not included in the output file. This
                          is desireable when trying to save space in the output and/or excluding all
                          files for security purposes. If code blocks are not included, an Info block
                          is still included in each log entry that displays the file name and line number
                          of where the log message was recorded.
            datetime_range: Defines the start and/or end date parameters of the log data included in the
                            output file. This is always a ``Tuple`` of size 2 where the first item is the
                            start date and the second is the end date. If only a start or end date is
                            desired, then set the other value to ``None``. Date should be ``datetime``
                            objects.
            exclude_files: A list of regular expressions for files that should NOT be included in the
                           code blocks. If ``include_code=True``, then the code files are included in
                           the output so the code can be referenced directly by the logs. This parameter
                           can be used to improve security and space. It is recommended to exclude files
                           that may leak sensitive information and are highly unlikely to provide much
                           use to the log file.
        """
        asyncio.run(self._generate(log_file=log_file, title=title, include_code=include_code,
                                   datetime_range=datetime_range, exclude_files=exclude_files))

    async def _generate(self, log_file: str, title: str, include_code: bool,
                        datetime_range: Tuple[Optional[datetime], Optional[datetime]],
                        exclude_files: List[str]):
        # region Get Content Of Log File
        log_file = Path(log_file)
        self._sql = LoggerSql(log_file)
        # endregion Get Content Of Log File

        # region Create HTML File
        log_tags = self._get_log_tags()

        log_entries_query = f'select * from {self._sql.log_entries.table_name}'
        if isinstance(datetime_range, (tuple, list)):
            if not len(datetime_range) == 2:
                raise ValueError(f'Datetime range must be a Tuple object of size 2 where the first '
                                 f'element, a datetime object, represents the "from" date and the '
                                 f'second, also a datetime object, represents the "to" date. To omit '
                                 f'one or the other, set the value to None.')
            dt_fmt = lambda x: x.strftime('%Y-%m-%d %H:%M:%S')
            if start := datetime_range[0]:
                if end := datetime_range[1]:
                    log_entries_query += f" where timestamp between '{dt_fmt(start)}' and '{dt_fmt(end)}'"
                else:
                    log_entries_query += f" where timestamp >= '{dt_fmt(start)}'"
            elif end := datetime_range[1]:
                log_entries_query += f" where timestamp <= '{dt_fmt(end)}'"
        log_entries = self._sql.select(query=log_entries_query)

        title = title or 'Log Results'
        code_blocks = self._get_code_blocks(log_entries=log_entries,
                                            exclude_files=exclude_files) if include_code else None
        # Minify HTML file to reduce space.
        html = htmlmin.minify(f"""
        <html>
            <head>
                <title>{title}</title>
                <link rel="stylesheet" type="text/css" href="./{Path(self.main_css_file).stem}.css" />
                <script type="text/javascript" src="./{Path(self.js_file).stem}.js"></script>
                <style>
                    {await self._get_styles(log_tags=log_tags, include_code=include_code)}
                </style>
            </head>
            <body>
                <!-- Title -->
                <div id="title">
                    <h1>{title.upper()}</h1>
                </div>
                <!-- Legend -->
                <div id="legend">
                    {await self._legend(log_tags=log_tags)}
                </div>
                <!-- Code Blocks -->
                <div class="hide" id="code-blocks">
                    {''.join([cb['block'] for cb in code_blocks.values()]) if code_blocks else ''}
                </div>
                <!-- Log Entries -->
                <div id="log-entries">
                    {await self._get_log_entries(log_tags=log_tags, log_entries=log_entries,
                                                 code_blocks=code_blocks if include_code else None)}
                </div>
            </body>
            {'<br/>' * 12}
        </html>
        """, remove_comments=True, remove_empty_space=True)
        # endregion Create HTML File

        # region Send Files To Log Directory
        resources = f'{log_file.parent}'
        file_path = ''
        for part in resources.split(os.sep):
            file_path += part + os.sep
            if not os.path.exists(file_path):
                os.mkdir(file_path)

        if code_blocks:
            for values in code_blocks.values():
                with open(f"{resources}/{values['path']}", 'w') as f:
                    f.write(values['code'])

        shutil.copy(
            src=self.python_css_file,
            dst=resources
        )
        shutil.copy(
            src=self.main_css_file,
            dst=resources
        )
        shutil.copy(
            src=self.js_file,
            dst=resources
        )
        with open(f'{log_file.parent}/{log_file.stem}.html', 'w') as f:
            f.write(html)
        # endregion Send Files To Log Directory

    def _get_log_tags(self):
        log_tags = self._sql.select(
            query=f'select * from {self._sql.log_tags.table_name}'
        )
        return {
            tag[self._sql.log_tags.name]: LogTag(
                name=tag[self._sql.log_tags.name],
                value=tag[self._sql.log_tags.value],
                html_color=tag[self._sql.log_tags.color]
            )
            for tag in log_tags.iterate()
        }

    @staticmethod
    async def _get_styles(log_tags: Dict[str, LogTag], include_code: bool):
        """
        Sets the global dynamic style variables and classes. These styles cannot be controlled
        by a stylesheet because they are dynamically created based on the log tags.
        """
        colors = []
        classes = []
        for ll in log_tags.values():
            colors.append(f'--{ll.alias}: {ll.html_color};')  # Color of log tag.
            # Each log tag has its own style class for highlighting the bottom of the
            # log entry.
            classes.append(f"""
            .{ll.alias} {{
                border-left: var(--{ll.alias}) solid 5px;
                border-bottom: lightgrey solid 2px;
                transition: border-bottom 0.2s ease-in-out;
            }}

            .{ll.alias}:hover {{
                border-bottom: var(--{ll.alias}) solid 2px;

            }}
            """)
        colors = '\n\t'.join(colors)
        classes = '\n'.join(classes)
        # Send the style classes along with a variable that sets the number of log entry buttons
        # to show. The "Code" button can be omitted.
        return f"""
        :root {{
            {colors}
            --log-entry-btns: {3 if include_code else 2}
        }}

        {classes}
        """

    @staticmethod
    async def _legend(log_tags: Dict[str, LogTag]):
        """
        Sets the log tags and the search filters in the legend.
        """
        def add_filter(log_tag: LogTag):
            return f"""
            <div class="log-tag fancy-checkbox" 
                 aria-label="{log_tag.alias}" 
                 aria-valuetext="{log_tag.value}"
                 onclick="handleLogTagFilter(this)">
                <label class="switch">
                    <input id="{log_tag.alias}" type="checkbox" checked>
                    <span 
                        class="slider round" 
                        style="background-color: {log_tag.html_color}; box-shadow: 0 0 10px {log_tag.html_color};">
                    </span>
                </label>
                <span>{log_tag.name}</span>
            </div>
            """

        filters = '\n'.join([
            add_filter(log_tag=log_tag)
            for log_tag in log_tags.values()
        ])
        return f"""
        <div id="filters-container">
            <div id="filters-controls">
                <button id="filters-btn" class="btn" onclick="handleShowFilter(this)">Hide Filters</button>
            </div>
            <div id="filters-wrapper">
                <div id="filters" class="decompressed">
                    <div id="log-tags">
                        {filters}
                    </div>
                    <div class="divider"></div>
                    <div id="search-container">
                        <div id="search-params">
                            <div class="search-param fancy-checkbox">
                                <label class="switch">
                                    <input id="search-in-msg" type="checkbox" checked>
                                    <span class="slider round"></span>
                                </label>
                                <span>Include Message Blocks</span>
                            </div>
                            <div class="search-param fancy-checkbox">
                                <label class="switch">
                                    <input id="search-in-info" type="checkbox">
                                    <span class="slider round"></span>
                                </label>
                                <span>Include Info Blocks</span>
                            </div>
                            <div></div>
                            <div class="search-param fancy-checkbox">
                                <label class="switch">
                                    <input id="search-whole-word" type="checkbox">
                                    <span class="slider round"></span>
                                </label>
                                <span>Whole Word Only</span>
                            </div>
                            <div class="search-param fancy-checkbox">
                                <label class="switch">
                                    <input id="search-regex" type="checkbox">
                                    <span class="slider round"></span>
                                </label>
                                <span>Use Regular Expression</span>
                            </div>
                            <div class="search-param fancy-checkbox">
                                <label class="switch">
                                    <input id="search-match-case" type="checkbox">
                                    <span class="slider round"></span>
                                </label>
                                <span>Match Case</span>
                            </div>
                        </div>
                        <div id="search-controls">
                            <input id="search" type="text" placeholder="Search for text..." onkeyup="searchOnEnter(event, this.id)">
                            <button id="go-btn" class="btn" onclick="search('search')">Go</button>
                        </div>
                        <div id="search-count"></div>
                    </div>
                </div>
            </div>
        </div>
        """

    def _get_code_blocks(self, log_entries: SelectResult, exclude_files: List[str] = None):
        """
        Creates the code block panel, code block object tags, and the file name for each block.
        The object tag uses a function to send a line number to scroll to when the frame is opened.
        """
        exclude_files = exclude_files or []
        cols = self._sql.log_entries
        codes = {}  # type: Dict
        fc = 0
        if exclude_files:
            exclude_regexes = "(" + ")|(".join(exclude_files) + ")"
        else:
            exclude_regexes = None
        for le in list(log_entries.iterate()):
            fp = le[cols.file_path]
            if fp not in codes.keys():
                if exclude_regexes and re.match(pattern=exclude_regexes, string=fp, flags=re.IGNORECASE):
                    continue
                with open(fp, 'r') as f:
                    code = f.read()
                code_html = highlight(
                    code=code,
                    lexer=PythonLexer(),
                    formatter=HtmlFormatter(linenos='inline')
                )
                fc += 1
                file_id = f"file-id-{fc}"
                # Minify to reduce space.
                code = htmlmin.minify(f"""
                <html>
                    <head>
                        <link rel="stylesheet" type="text/css" href="{self.python_css_file.name}">
                        <script>
                            window.addEventListener("message", function(event) {{
                                line_num = Number(event.data);
                                view_line = line_num > 2 ? line_num - 3 : 0;
                                linenos = document.querySelectorAll('.lineno');
                                if(linenos.length > 0) {{
                                    linenos.forEach(line => {{line.style.backgroundColor = 'transparent';}})
                                    linenos[line_num - 1].style.backgroundColor = 'lightgreen';
                                    body = document.querySelector('body');
                                    body.scrollTo({{top: linenos[view_line].offsetTop, behavior: 'smooth'}});
                                }}
                            }})
                        </script>
                    </head>
                    <body>
                        {code_html}
                    </body>
                </html>
                """, remove_empty_space=True)
                # A UUID is used on the filename to create a consistent but unique filename for the HTML file
                # for the code. This prevents clashing of filenames.
                path = f'{uuid.uuid3(uuid.NAMESPACE_OID, fp).hex}.html'
                codes[fp] = {
                    'id'   : file_id,
                    'path' : path,
                    'block': f'''
                        <div class="code-block hide" id="{file_id}" aria-label="{Path(fp).stem}.html">
                            <div class="code-block-controls">
                                <span class="filepath" title="{le[cols.file_path]}">{le[cols.file_path]}</span>
                                <button class="close-code" onclick="hideCode('{file_id}')">X</button>
                            </div>
                            <div>
                                <object class="code" data="./{path}" type="text/html"></object>
                            </div>
                        </div>
                    ''',
                    'code' : code
                }
        return codes

    async def _get_log_entries(self, log_tags: Dict[str, LogTag], log_entries: SelectResult,
                               code_blocks: Dict[str, Dict[str, str]]):
        """
        Creates the reset button, log entries, and log blocks.
        """
        cols = self._sql.log_entries

        file_names = []

        def format_info_block(le: dict):
            return f'File: {le[cols.file_path]}\n' \
                   f'Line Number: {le[cols.line_num]}\n' \
                   f'Qualified Name: {le[cols.function_name]}\n' \
                   f'Thread Id: {le[cols.thread_id]}\n' \
                   f'Thread Name: {le[cols.thread_name]}\n' \
                   f'Timestamp: {le[cols.timestamp]}\n' \
                   f'Log Tag: {le[cols.tag_name]}'

        logs = [
            f"""
            <div id="log-entries-controls">
                <div class="btn">
                    <button id="reset-btn" onclick="resetLogBlocks()">Reset</button>
                </div>
            </div>
            """
        ]

        def organize(les: List[dict]):
            """
            Organizes the log entries by thread and timestamp.
            """
            results = []
            threads = {}  # type: Dict[str, Tuple[int, List[Dict]]]
            main_counter = 0
            for le in les:
                if bool(le[cols.is_main_thread]):
                    results.append(le)
                    main_counter += 1
                else:
                    if (thread_id := le[cols.thread_id]) not in threads.keys():
                        threads[thread_id] = (main_counter, [le])
                    else:
                        threads[thread_id][1].append(le)
            for thread_id, value in threads.items():
                index, entries = value
                original = results.copy()
                new = original[:index]
                new.extend(entries)
                new.extend(original[index:])
                results = new

            return results

        log_entries = organize(les=list(log_entries.iterate()).copy())

        for e, log_entry in enumerate(log_entries):
            log_tag = log_tags[log_entry[cols.tag_name]]
            log_entry_id = f'log-entry-{e}'
            msg_block_id = f'msg-block-{e}'
            code_block_id = f'code-block-{e}'
            info_block_id = f'info-block-{e}'
            log_group_id = f'log-group-{e}'

            display = "hide" if log_entry[cols.depth] > 0 else "show"
            # Only show exp btn if there are child entries to this entry.
            if e < (len(log_entries) - 1) and \
                    log_entries[e + 1][cols.depth] > log_entry[cols.depth]:
                expand_btn = f"""
                <div class="btn item-container">
                    <button class="exp-btn hide-logs" onclick="toggleExpLogs('{log_entry_id}')"></button>
                </div>
                """
            else:
                expand_btn = f"""
                <div class="item-container">
                    <div class="no-exp">-</div>
                </div>
                """
            # Only show the code block button if enabled.
            if code_blocks:
                code_block = code_blocks.get(log_entry[cols.file_path])
                fid, fp = code_block.get('id'), code_block.get('path')
                code_btn = f'''
                    <div class="code-btn btn item-container">
                        <button onclick="showCode('{fid}', '{fp}', {log_entry[cols.line_num]})">Code</button>
                    </div>
                '''
            else:
                code_btn = ''
            func_def = f'{log_entry[cols.function_name]}:{log_entry[cols.line_num]}'
            logs.append(f"""
            <div id='{log_entry_id}' 
                 class="log-entry {display}"
                 aria-valuetext="{log_tag.alias}" 
                 aria-label="{log_entry[cols.depth]}">
                <div class="log-row {log_tag.alias}" style="margin-left: {log_entry[cols.depth] * 25}px;">
                    {expand_btn}
                    <div class="func item-container">
                        <span title="{func_def}">{func_def}</span>
                    </div>
                    <div class="preview item-container">
                        <span>{log_entry[cols.msg]}</span>
                    </div>
                    <div class="msg-btn btn item-container">
                        <button onclick="handleLogContent(this, '{msg_block_id}')">Message</button>
                    </div>
                    <div class="info-btn btn item-container">
                        <button onclick="handleLogContent(this, '{info_block_id}')">Info</button>
                    </div>
                    {code_btn}
                </div>
                <div 
                    class="log-block-container" 
                    style="border-left: var(--{log_tag.alias}) solid 5px; margin-left: {log_entry[cols.depth] * 25}px;"
                >
                    <div id="info-block-{e}" class="info-block log-block hide">
                        <pre>{format_info_block(log_entry)}</pre>
                    </div>
                    <div id="msg-block-{e}" class="msg-block log-block hide">
                        <pre>{log_entry[cols.msg]}</pre>
                    </div>
                </div>
            </div>
            """)
            e += 1
        return ''.join(logs)



from typing import List, Callable, Any, Union
import os
import sys
import dis
import traceback
import inspect
from datetime import datetime
import jsonpickle
import re
import threading
from pathlib import Path
from contextlib import contextmanager
from dblogging.config import LogTag, LogTags
from dblogging.generators import HtmlLogGenerator
from dblogging.sqlite.dal import LoggerSql


class _LogThread:
    def __init__(self, depth: int = -1, sql: LoggerSql = None, blacklist_function: Callable = None,
                 mode: str = 'all'):
        self.depth = depth
        self.sql = sql
        self._mode = mode
        self.persistence_enabled = mode in {'all', 'persistence'}
        self.console_enabled = mode in {'all', 'console'}
        self.blacklist_function = blacklist_function

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value: str):
        if value == 'current':
            return
        elif value not in (values := {'all', 'persistence', 'console'}):
            raise ValueError(f'Cannot set mode to "{value}" because it is invalid.\n'
                             f'Valid modes are {values}')
        self._mode = value
        self.persistence_enabled = value in {'all', 'persistence'}
        self.console_enabled = value in {'all', 'console'}


def singleton(cls):
    instances = {}

    def _singleton(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return _singleton


@singleton
class Logger:
    def __init__(self):
        """
        Creates a singleton instance of the logger. This object is a singleton object because different
        instances of the logger can create issues if persisting to the same file. Rather than having
        multiple instances of the logger, a single instance can be used while using log tags.

        --> Log Tags

        Log tags are a conceptual representation of how the program can be divided in terms of logging. Common
        tags are severity tags: info, debug, warning, error, and critical. However, custom tags can be created
        to represent the logging in more and different ways. Rather than severity tags, tags could alternatively
        represent the level of code that is performing the log, such as the Data Access Layer vs. a public API
        layer. Log tags provide a way to tag log messages so they can be more easily filtered when viewing the
        output.

        To create custom log tags, either import the LogTagTemplate or the LogTags and inherit it in the new
        custom log tag class. A ``default`` and ``critical`` log tags are required because the logger uses
        defaults for standard and error logging. The LogTagTemplate does not define any of the required tags,
        but raises an error when they are not implemented whereas the LogTags does. Here is an example:

            from logger.config import LogTagTemplate, LogTags, LogTag

            # Inheriting from LogTagTemplate requires default and critical tags to be defined.
            class CustomLogTags1(LogTagTemplate):
                default = LogTag(
                    name='Standard',
                    value=0,
                    html_color='cyan'
                )

                test = LogTag(
                    name='Test',
                    value=10,
                    html_color='green'
                )

                critical = LogTag(
                    name='Critical',
                    value=90,
                    html_color='red'
                )

            # Inheriting from LogTags doesn't require default and critical tags to be defined.
            class CustomLogTags2(LogTags):
                test = LogTag(
                    name='Test',
                    value=10,
                    html_color='green'
                )

        To use the custom log tags, set the ``log_tags`` to the class that defines them. Using the
        example above, ``logger.log_tags = CustomLogTags1``. Then ``logger.log(msg, CustomLogTags1.test)``
        would be legal.

        --> Basic Logging

        There are three basic logging methods: ``log()``, ``log_method()``, and ``log_exception()``.

        ``log()`` simply commits a log message on the fly. The function name is retrieved dynamically.
        ``log_method()`` should at most rarely be used. It creates a log entry with a given function name.
        ``log_exception()`` is used to locate the exception in the call stack to log.

        --> Logging Control

        Logging output can be controlled by setting rules once the ``start()`` method is called. These
        rules are set using the ``set_rule()`` method. These rules control a few factors when logging:
            a) The mode. There are four modes: current, console, persistence, and all. The console mode
               only logs to stdout. The persistence mode only logs to the DB log file. All does both. The
               current mode does not switch the mode when setting the rule.
            b) The rule. Basic rules are setting a minimum tag value or blacklisting a set of tag names.
               A custom rule can be supplied as well. This controls what will and won't get logged to
               reduce noise in the logs.
            c) Why. This is a message suggesting why the rule was changed. This will always be logged.

        --> Contextual Logging

        Use ``generate()`` as a context manager (i.e. ``with logger.generate('html'):``). Nothing is yielded
        to the context block. After the context is executed, the logger automatically searches the call stack
        if an error is raised and logs it. If persistence is enabled, the format generator's ``generate()``
        method is called to create the log file from the DB file. HTML is the only generator provided, although
        a custom generator can be created.

        --> Log Wrappers

        There are two log wrappers: wrap_class() and wrap_func(). They log the inputs and outputs of functions.
        They provided depth to the log output, meaning a hierarchy of logs can be created. Without these methods
        all logs would be flat and one dimensional. See their docstrings for more details.
        """
        self._disabled = True
        self._date_format = '%Y/%m/%d %H:%M:%S'
        self._timestamp = lambda: datetime.now().strftime(self._date_format)
        self._persistent_logging = False
        self._log_path = None
        self._log_tags = LogTags
        self._sql = None
        self._main_thread = threading.current_thread()
        self._threads = {self._main_thread.ident: _LogThread(depth=0)}
        self._jsonpickle = jsonpickle
        self._jsonpickle.set_encoder_options('json', sort_keys=True, indent=4)
        self.logger_lock = threading.Lock()

    @property
    def log_path(self) -> Path:
        return self._log_path

    @log_path.setter
    def log_path(self, path: Union[Path, str]):
        if self._log_path and os.path.exists(self._log_path):
            raise ValueError(f'Cannot set log path to "{path}" because it is already set to '
                             f'"{self._log_path}", which exists.')
        if not isinstance(path, Path) and not path.endswith('.db'):
            path += '.db'
            path = Path(path)
        self._log_path = path
        self._persistent_logging = True

    @property
    def log_tags(self):
        return self._log_tags

    @log_tags.setter
    def log_tags(self, value):
        if not hasattr(value, 'default'):
            raise AttributeError(f'The custom logger must have a <default> LogTag attribute defined.')
        elif not hasattr(value, 'critical'):
            raise AttributeError(f'The custom logger must have a <critical> LogTag attribute defined.')
        self._log_tags = value

    @property
    def date_format(self):
        return self._date_format

    @date_format.setter
    def date_format(self, format: str):
        self._date_format = format

    def start(self):
        """
        Starts the logging. By default all logging is disabled. Once this method is called, logging
        is initialized according to the log path, log tags, and date format set. By default, nothing
        is set. These variables can be changed throughout the execution of the program, although it
        is not recommended as the output can produce unexpected results.

        If nothing is set and this method is called, only stdout is enabled. No logs will persist.

        The ``log_path`` can change throughout the execution of the program given that the file does
        not already exist (but the parent folder must exist). Once set, call ``start()`` again to
        reinitialize the DB file.

        The ``log_tags`` can change, although this is dangerous and should not happen. Once this
        method is called, the DB is initialized with all of the log tag data. If the log tags
        change during execution of the program and contains tags that did not exist previously,
        then they will not be included in the output logs.

        The ``date_format`` can change any time as this only affects stdout.
        """
        self._disabled = False
        if self.log_path:
            self._sql = LoggerSql(db_file_name=self._log_path)
            self._threads = {
                self._main_thread.ident: _LogThread(
                    depth=0,
                )
            }

            self._sql.create_database()
            for ll in self._log_tags.get_all():
                self._sql.log_tags.insert(
                    name=ll.name, value=ll.value, color=ll.html_color
                )

    @contextmanager
    def generate(self, format_generator: Union[str, Callable], **kwargs):
        """
        Generates logs and optionally generates an output file from the DB log file according
        to the ``format_generator`` parameter. The ``format_generator`` is either a string
        that identifies a built-in format generator or a custom generator that is callable.

        The only built-in generator is 'html' at this time.

        A custom generator can be used. It must be able to be initialized without parameters
        and have a ``generate()`` function with at least a ``log_file`` parameter. Optional
        parameters can be passed to the generator as ``kwargs``. For example:

            # Using the built-in HTML generator.
            with logger.generate('html'):
                # Do stuff here.

            # Using a custom generator.
            class SomeCustomGenerator:
                def __init__(self):  # No arguments required!
                    # Initialize stuff here.

                # This method is required!
                def generate(log_file: str, **kwargs): # log_file required!
                    # Generate output here.

            with logger.generate(SomeCustomGenerator, custom_kw='Some Custom Variable Value'):
                # Do stuff here.

        Args:
            format_generator: Either a string for a built-in generator or a callable custom one.
            **kwargs: Extra arguments to pass to the generator.
        """
        if self._disabled:
            self.start()
        if self._persistent_logging:
            if isinstance(format_generator, str):
                generators = {
                'html': HtmlLogGenerator
                }
                generator = generators[format_generator]()
            else:
                generator = format_generator()
            if not self._log_path:
                raise ValueError('Cannot generate persisting logs without a log path.\n'
                                 'Please instantiate the logger with a log path.')
            try:
                if not os.path.exists(self._log_path.parent):
                    file_path = ''
                    for path in self._log_path.parts:
                        file_path += path + os.sep
                        if not os.path.exists(file_path):
                            os.mkdir(file_path)
                yield
            except:
                self.log_exception()
                raise
            finally:
                generator.generate(log_file=self._log_path, **kwargs)
        else:
            try:
                yield
            except:
                self.log_exception()

    def set_rule(self, mode: str = 'current', log_tag: LogTag = None, min_tag_value: Union[LogTag, int] = None,
                 blacklist_tag_names: List[str] = None, blacklist_function: Callable = None,
                 reset: bool = False, why: str = ''):
        """
        Sets a rule that dictates what will and will not get logged, as well as how they are logged.

        The log mode can be used to change where the logs are recorded. This is useful for simply
        reducing noise in the log output.
        There are four modes: current, console, persistence, and all.
            * current: Do not change the mode. Default.
            * console: Change the mode to log to the console only.
            * persistence: Change the mode to log to the SQLite DB file only.
            * all: Change the mode to use both console and persistent logging.

        The ``min_tag_value`` specifies the minimum tag value that a log tag must have in order to
        log a message.

        The ``blacklist_tag_names`` specifies a list of tag names that prevent messages from being
        logged. In other words, a log tag whose name is blacklisted will not log the message.

        The ``blacklist_function`` defines a custom function that blacklist log tags from logging
        their messages. There is no whitelist option. The function must return a boolean where
        ``False`` allows a message to be logged.

        The ``reset`` flag resets the log rules to their initial state when the logger was
        initialized.

        If none of ``min_tag_value``, ``blacklist_tag_names``, ``blacklist_function``, or ``reset`` are
        specified, then no rules are set to restrict any logging. In other words, all logging in the
        specified mode is enabled.

        Args:
            mode: One of 'current', 'console', 'persistence', or 'all'. Default is 'current'. Case sensitive.
            log_tag: The log tag that logs the message concerning the rule change. Defaults to the default tag.
            blacklist_function: Optional. This is evaluate first. If not ``None``, this rule is used.
            min_tag_value: Optional. This is evaluated second. If not ``None``, this rule is used. Can be either
                           the log tag, the log tag value, or any integer.
            blacklist_tag_names: Optional. This is evaluated third. If not ``None``, this rule is used. Must
                                 be a list of tag names.
            reset: Optional. This is evaluated last. If not ``None``, the rule in the specified mode is reset
                   to the initial state.
            why: A description of why the rule is being set.
        """
        if self._disabled:
            return
        log_tag = log_tag or self._log_tags.default
        msg = f'Setting log mode to "{mode}". '
        if blacklist_function:
            msg += f'A function has been defined to decide what is logged.'
        elif min_tag_value:
            if isinstance(min_tag_value, LogTag):
                min_tag_value = min_tag_value.value
            blacklist_function = lambda x: x.value < min_tag_value
            msg += f'Only tag values greater than or equal to {min_tag_value} will be logged.'
        elif blacklist_tag_names:
            blacklist_function = lambda x: x.name in blacklist_tag_names
            msg += f'These tag names are will not be logged: {blacklist_tag_names}'
        elif reset:
            blacklist_function = lambda x: False
            msg += f'Resetting rule. There are no restrictions to logging in this rule.'
        else:
            blacklist_function = lambda x: False
            msg += f'No rule defined. There are no restrictions to logging in this rule.'

        func, file_path, line_num = self._get_log_info(num_prev_callers=0)
        self._commit_log(msg=f'{msg}\n{why}', log_tag=log_tag, file_path=file_path, line_num=line_num, func=func,
                         force=True)

        thread = threading.current_thread()
        if thread.ident == self._main_thread.ident:
            for thread in self._threads.values():
                thread.mode = mode
                thread.blacklist_function = blacklist_function
        elif lt := self._threads.get(thread.ident):
            lt.mode = mode
            lt.blacklist_function = blacklist_function
        else:
            self._threads[thread.ident] = _LogThread(
                depth=self._threads[self._main_thread.ident].depth,
                blacklist_function=blacklist_function,
                mode=mode
            )

    def wrap_class(self, log_tag: LogTag = None, func_regex_exclude: str = '', mask_input_regexes: List = None,
                   mask_output: bool = False):
        """
        Applies ``wrap_func()`` to each callable member of a class that does not start with "__" and does not
        match the ``func_regex_exclude``. If a callable member starting with "--" should be wrapped, then it
        must be explicitly wrapped with the ``wrap_func()`` method. The ``mask_input_regexes`` and ``mask_output``
        parameters are passed to the ``wrap_func()`` method.

        Args:
            log_tag: The log tag to apply to each method wrapped.
            func_regex_exclude: A regular expression matching method names that should not be wrapped.
            mask_input_regexes: Passed to ``wrap_func()``.
            mask_output: Passed to ``wrap_func()``.
        """
        log_tag = log_tag or self._log_tags.default

        def _wrap(cls):
            for attr, fn in inspect.getmembers(cls, inspect.isroutine):
                if callable(getattr(cls, attr)) and not fn.__name__.startswith('__'):
                    if func_regex_exclude:
                        matches = re.findall(pattern=func_regex_exclude, string=fn.__name__, flags=re.IGNORECASE)
                        if fn.__name__ in matches:
                            continue

                    if type(cls.__dict__.get(fn.__name__)) in {staticmethod, classmethod}:
                        setattr(cls, attr, self.wrap_func(
                            log_tag=log_tag,
                            is_static_or_classmethod=True,
                            mask_input_regexes=mask_input_regexes,
                            mask_output=mask_output
                        )(getattr(cls, attr)))
                    else:
                        setattr(cls, attr, self.wrap_func(
                            log_tag=log_tag,
                            mask_input_regexes=mask_input_regexes,
                            mask_output=mask_output
                        )(getattr(cls, attr)))
            return cls

        return _wrap

    def wrap_func(self, log_tag: LogTag = None, mask_input_regexes: List = None, mask_output: bool = False,
                  is_static_or_classmethod: bool = False):
        """
        Wrapper for a function or method. This wrapper performs a few important steps:
            1. Increases the depth of the call, which controls its relation to all previous and
               subsequent logs in the log hierarchy.
            2. Logs the function's name, inputs, and outputs.
            3. Catches any exceptions raised by the function and logs them.
            4. Decreases the depth of the call to return to the original depth of the call hierarchy.

        The depth is the defining marker for the position of a function in the logged hierarchy. Without it,
        all logs would be flat an hard to parse through.

        Example:

            @logger.wrap_func(
                log_tag=Tags.SomeTag,
                mask_input_regexes=['cls', 'password'],
                is_static_or_classmethod=True
            )
            @classmethod
            def do_something(cls, username, password):
                # Do some stuff
                return {'API Token': 'xyz'}

        * The above example will use the ``Tags.SomeTag`` tag to log the calling and the returning message.
        * The ``cls`` and ``password`` parameters will be masked. The message will show this:

          Called do_something
          Arguments:
          {
              "cls": "********",
              "username": "Admin",
              "password": "********"
          }

        * The output is not masked, so the output will show this:

          do_something returned.
          Return Values:
          {
              "API Token": "xyz"
          }

        * If ``is_static_or_classmethod`` was not set to ``True``, then this wrapper would have failed.

        Args:
            log_tag: The log tag to apply to the input and output log messages.
            mask_input_regexes: A list of regular expressions matching input variable names.
            mask_output: When ``True`` the output is masked like this: "********".
            is_static_or_classmethod: Set this to ``True`` when wrapping a staticmethod or classmethod.
        """
        log_tag = log_tag or self._log_tags.default

        def __wrap(func):
            def __wrapper(*args, **kwargs):
                if self._disabled:
                    return func(*args, **kwargs)

                thread = threading.current_thread()
                # If the current thread hasn't been accounted for, account for it by creating a
                # _LogThread using the main thread's configuration.
                if not self._threads.get(thread.ident):
                    mt = self._threads.get(self._main_thread.ident, _LogThread())
                    self._threads[thread.ident] = _LogThread(
                        depth=mt.depth,
                        blacklist_function=mt.blacklist_function,
                        mode=mt.mode
                    )

                # Before the function is called.
                try:
                    if is_static_or_classmethod:
                        # staticmethod and classmethod objects inject another argument at index 0.
                        args = args[1:]
                    params = dict(inspect.signature(func).bind(*args, **kwargs).arguments)  # type: dict
                except TypeError as e:
                    self.log(
                        msg='\n'.join(e.args),
                        log_tag=self._log_tags.critical,
                        num_prev_callers=2
                    )
                    raise TypeError(e)

                before_string = 'Called ' + func.__qualname__
                if params:
                    if mask_input_regexes:
                        in_regexes = "(" + ")|(".join(mask_input_regexes) + ")"
                        for key in params.keys():
                            if re.match(pattern=in_regexes, string=key, flags=re.IGNORECASE):
                                params[key] = '********'
                    before_string += '\nArguments:\n' + self._jsonpickle.dumps(params, max_depth=3, unpicklable=False)
                self.log_method(func=func, msg=before_string, log_tag=log_tag, returning=False)
                self._threads[thread.ident].depth += 1
                try:
                    result = func(*args, **kwargs)

                    # After the function returns.
                    after_string = f'{func.__qualname__} returned.'
                    if result is not None:
                        if mask_output:
                            after_string += ' Output is masked.'
                        else:
                            ret_vals = self._jsonpickle.dumps(result, max_depth=3, unpicklable=False)
                            after_string += f'\nReturn Values: {ret_vals}'
                    self.log_method(func=func, msg=after_string, log_tag=log_tag, returning=True)

                    return result
                except:
                    self.log_exception()
                    raise
                finally:
                    self._threads[thread.ident].depth -= 1

            return __wrapper

        return __wrap

    def _commit_log(self, msg: Any, log_tag: LogTag, file_path: str, line_num: int, func: str = None,
                    force: bool = False):
        """
        Commits the log transaction to the console and the SQLite DB.

        Args:
            msg: Log message.
            log_tag: The log tag that logs the message. Defaults to the default tag.
            file_path: Absolute path to the file in relation to the log message.
            line_num: Line number in the ``file_path``.
            func: Optional. The qualified name of the function in relation to the ``line_num``.
            force: Optional. If ``True``, then the log thread settings are ignored and the message is logged to
                   the console and, if ``persistent logging==True``, then the message is persisted.
        """
        thread = threading.current_thread()
        # If the current thread hasn't been accounted for, account for it by creating a
        # _LogThread using the main thread's configuration.
        if (mt := self._threads.get(thread.ident)) is None:
            mt = self._threads.get(self._main_thread.ident, _LogThread())
            self._threads[thread.ident] = _LogThread(
                depth=mt.depth,
                blacklist_function=mt.blacklist_function,
                mode=mt.mode
            )

        if mt.console_enabled or force:
            print(f'[{log_tag.name}] {self._timestamp()}: {msg}')

        if self._persistent_logging and (mt.persistence_enabled or force):
            with self.logger_lock:
                self._sql.log_entries.insert(
                    file_path=file_path,
                    function_name=func,
                    line_num=line_num,
                    msg=str(msg),
                    tag_name=log_tag.name,
                    depth=self._threads[thread.ident].depth,
                    thread_id=thread.ident,
                    thread_name=thread.name,
                    is_main_thread=(thread.ident == self._main_thread.ident)
                )

    def _get_log_info(self, num_prev_callers: int):
        with self.logger_lock:
            frame = inspect.currentframe()
        # Add two because this should only be called once internally.
        outer_frames = inspect.getouterframes(frame)[num_prev_callers + 2]
        if outer_frames.function == '<module>':
            func = inspect.getmodulename(outer_frames.filename)
        else:
            if (class_name := outer_frames[0].f_locals.get('self')) is not None:
                func = f'{class_name.__class__.__qualname__}.{outer_frames.function}'
            else:
                func = outer_frames.function
        return func, outer_frames[1], outer_frames[2]

    def log(self, msg: Any, log_tag: LogTag = None, num_prev_callers: int = 0):
        """
        Logs a message if logging is enabled and it satisfies the rule set. The default rule allows
        all logging. The rule may specify console logging only, persistent logging only, or both.

        ``num_prev_callers`` allows the caller to specify which caller in the call stack should be
        referenced in the log entry. By default the caller of this method is captured and logged.
        This is handy to use when the caller of this method doesn't live in a method whose context
        is helpful to debugging. For example, if A calls B and B calls ``log(..., prev_num_callers=1)``
        then the message is logged with a reference to the file path, line number, and function name
        from caller A in relation to the call stack.

        Args:
            msg: Log message.
            log_tag: The log tag that logs the message. Defaults to the default tag.
            num_prev_callers: Number of callers in the current call stack previous to the caller of
                              this method.
        """
        if self._disabled:
            return
        log_tag = log_tag or self._log_tags.default
        thread = threading.current_thread()
        if lt := self._threads.get(thread.ident):
            # Check the log rule. If it doesn't pass, return.
            if isinstance(lt.blacklist_function, Callable) and lt.blacklist_function(log_tag) is True:
                return

        func, file_path, line_num = self._get_log_info(num_prev_callers=num_prev_callers)
        self._commit_log(msg=msg, log_tag=log_tag, file_path=file_path, line_num=line_num, func=func)

    def log_exception(self):
        """
        Logs the function name, file path, line number, and exception message if logging is enabled. The
        log tag used is ``critical``. The log rule is disregarded.
        """
        if self._disabled:
            return
        with self.logger_lock:
            tb = sys.exc_info()[2]

        while True:
            if tb.tb_next is None:
                break
            tb = tb.tb_next
        frame = tb.tb_frame
        outer_frames = inspect.getouterframes(frame)[0]
        if outer_frames.function == '<module>':
            func = inspect.getmodulename(outer_frames.filename)
        else:
            if (class_name := outer_frames[0].f_locals.get('self')) is not None:
                func = f'{class_name.__class__.__qualname__}.{outer_frames.function}'
            else:
                func = outer_frames.function
        msg = traceback.format_exc()
        self._commit_log(msg=msg, log_tag=self._log_tags.critical, file_path=outer_frames[1], line_num=outer_frames[2],
                         func=func)

    def log_method(self, func: callable, msg: Any, log_tag: LogTag = None, returning: bool = False):
        """
        Similar to ``log()`` with one particular difference. ``log()`` dynamically retrieves the
        information of the referenced caller using ``num_prev_callers`` whereas ``log_method()``
        directly uses ``func`` for the name of the function and ``returning`` to decide if the
        line number referenced is the first or last line of the function. This method should rarely
        be used. To log a method use ``wrap_func()`` instead as it handles log depth.

        Args:
            func: A callable function.
            msg: Log message.
            log_tag: The log tag that logs the message. Defaults to the default tag.
            returning: If ``True``, reference the last line of the method, otherwise the first.
        """
        if self._disabled:
            return
        log_tag = log_tag or self._log_tags.default
        thread = threading.current_thread()
        if lt := self._threads.get(thread.ident):
            # Check the log rule. If it doesn't pass, return.
            if isinstance(lt.blacklist_function, Callable) and lt.blacklist_function(log_tag) is True:
                return

        self._commit_log(
            msg=msg,
            log_tag=log_tag,
            file_path=func.__code__.co_filename,
            line_num=list(dis.findlinestarts(func.__code__))[-1][-1] if returning else func.__code__.co_firstlineno,
            func=func.__qualname__
        )
